/*
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * Copyright (C) 2000-2006 HighPoint Technologies, Inc.
 */
#include "global.h"
#include "hptintf.h"
#include "oslinux.h"

#if DBG
#ifdef DEBUG_LEVEL
int hpt_dbg_level = DEBUG_LEVEL;
#else 
int hpt_dbg_level = 0;
#endif
#endif

MODULE_AUTHOR ("HighPoint Technologies, Inc.");
MODULE_DESCRIPTION ("HPT374 Controller driver");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include <linux/reboot.h>
#endif

#ifndef DECLARE_MUTEX_LOCKED
#define DECLARE_MUTEX_LOCKED(sem) struct semaphore sem = __SEMAPHORE_INITIALIZER(sem, 0)
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define GLOBAL_DATA_ALLOC(size) \
	({ void *p = kmalloc(size, GFP_ATOMIC); if (p) memset(p, 0, size); p; })
#define GLOBAL_DATA_FREE(p, size) kfree(p)
#else 
#define GLOBAL_DATA_ALLOC(size) scsi_init_malloc(size, GFP_ATOMIC)
#define GLOBAL_DATA_FREE(p, size) scsi_init_free((char*)p, size)
#endif

/* vmalloc have problem on some SMP kernels? */
#define VMALLOC(size) kmalloc(size, GFP_ATOMIC)
#define VFREE(p) kfree(p)

#include "glbdata.c"

typedef struct st_HPT_DPC {
	void (*dpc)(void *, UCHAR);
	void *arg;
	UCHAR flags;
} ST_HPT_DPC;

#define MAX_DPC 16
atomic_t DPC_Request_Nums = ATOMIC_INIT(0);
static spinlock_t DpcQueueLock = SPIN_LOCK_UNLOCKED;

static ST_HPT_DPC DpcQueue[MAX_DPC];
static int DpcQueue_First=0;
static int DpcQueue_Last = 0;

struct Scsi_Host *hpt_vhost = 0;
static int irq_map[MAX_CHIPS];


DECLARE_MUTEX_LOCKED(hpt_notify_wt);
struct semaphore * hpt_queue_ready = 0;
DECLARE_MUTEX_LOCKED(hpt_queue_wt);
struct task_struct *hpt_worker_thread_handle = 0; /* our kernel thread */
spinlock_t driver_lock = SPIN_LOCK_UNLOCKED;

PCommand pCommandBlocks=0;
PUCHAR SglBuffer[MAX_CHANNELS] = {0, };

volatile int stop_thread = 0;

void Check_Idle_Call(void);
void hpt_worker_thread(void *arg);
int hpt3xx_ProcInfo(char *buffer,char **start,off_t offset,int length, int hostno,int inout);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
int hpt3xx_ProcInfo26(struct Scsi_Host *, char *, char **, off_t, int, int);
#endif
void HPTLIBAPI OsSendCommand (Scsi_Cmnd *SCpnt);
void HPTLIBAPI fOsCommandDone(PCommand pCmd);
static void SetInquiryData(PINQUIRYDATA inquiryData, PVDevice pVDev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)   /* 2.2.x */
extern struct proc_dir_entry proc_scsi_hpt3xx;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
/* notifier block to get a notify on system shutdown/halt/reboot */
static int hpt_halt(struct notifier_block *nb, ulong event, void *buf);
static struct notifier_block hpt_notifier = {
    hpt_halt, NULL, 0
};
#endif

static ULONG LastStamp = 0;
ULONG HPTLIBAPI GetStamp(void)
{
	ULONG stamp;
	stamp =	jiffies;
	if(stamp == LastStamp)
		stamp++;
	LastStamp = stamp;
	return LastStamp;
}

/****************************************************************
 *	Name:	Irq_Handler
 *	Description: interrupt processing
 ****************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static void Irq_Handler (int irq, void *dev_id, struct pt_regs *regs)
{
    unsigned long flags, f2;
	UINT  i;
	PVBus pVBus;
	BOOLEAN ret = 0;

	KdPrintE(("----- Entering Isr() -----\n"));
    spin_lock_irqsave (&io_request_lock, flags);
    spin_lock_irqsave(&driver_lock, f2);

	FOR_EACH_VBUS(pVBus) {
		for(i = 0; i < pVBus->nInstances; i++){
		   if(HOSTDATA(shpnt)->irq[i] != irq) continue;
		   ret |= fChipInterrupt(pVBus->pChipInstance[i]);
		}
	}
	if(ret) {
		CheckPendingCall();
		if(CURRCMDs==0 && (atomic_read(&DPC_Request_Nums)==0)) Check_Idle_Call();	
	}
	spin_unlock_irqrestore(&driver_lock, f2);
    spin_unlock_irqrestore (&io_request_lock, flags);
	KdPrintE(("----- Leaving Isr() -----\n"));
}
#else 
static irqreturn_t Irq_Handler (int irq, void *dev_id
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
, struct pt_regs *regs
#endif
)
{
	unsigned long flags, f2;
	UINT  i;
	PVBus pVBus;
	BOOLEAN ret = 0;

	KdPrintE(("----- Entering Isr() -----\n"));
	spin_lock_irqsave (hpt_vhost->host_lock, flags);
    spin_lock_irqsave(&driver_lock, f2);

	FOR_EACH_VBUS(pVBus) {
		for(i = 0; i < pVBus->nInstances; i++){
		   if(HOSTDATA(shpnt)->irq[i] != irq) continue;
		   ret |= fChipInterrupt(pVBus->pChipInstance[i]);
		}
	}
	if(ret) {
		CheckPendingCall();
		if(CURRCMDs==0 && (atomic_read(&DPC_Request_Nums)==0)) Check_Idle_Call(_VBUS_P0);	
	}

	spin_unlock_irqrestore(&driver_lock, f2);
	spin_unlock_irqrestore(hpt_vhost->host_lock, flags);
	KdPrintE(("----- Leaving Isr() -----\n"));
	return IRQ_RETVAL(ret);
}
#endif
void Release_SC(Scsi_Cmnd**,Scsi_Cmnd*);
/* call this function inside driver_lock */
void Check_Idle_Call(void)
{
    if (mWaitingForIdle(gVBus)) {
		CheckIdleCall();
        up(&hpt_queue_wt);
	}
    /*check pending_Q*/
        while(PENDING_Q != NULL)
		{
			Scsi_Cmnd* SCpnt = (Scsi_Cmnd *)HPT_SCP(PENDING_Q)->ptr;
           		Release_SC(&PENDING_Q,SCpnt);
			CallAfterReturn((DPC_PROC)OsSendCommand, SCpnt);
		}
}

void Queue_SC(Scsi_Cmnd **SC_Q,Scsi_Cmnd *SCpnt)
{
	if(*SC_Q==NULL)
		HPT_SCP(SCpnt)->ptr = (char *)SCpnt;
	else
	{
		HPT_SCP(SCpnt)->ptr = HPT_SCP(*SC_Q)->ptr;
		HPT_SCP(*SC_Q)->ptr = (char *)SCpnt;
	}
	*SC_Q = SCpnt;
}

void Release_SC(Scsi_Cmnd **SC_Q,Scsi_Cmnd *SCpnt)
{
	Scsi_Cmnd *TmpSCpnt;
    TmpSCpnt = *SC_Q;
	
	if(HPT_SCP(SCpnt)->ptr == (char *)SCpnt)/*it means SCpnt is the last one in CURRCMDs*/
		*SC_Q = NULL;
	else
	{
		while(HPT_SCP(TmpSCpnt)->ptr != (char *)SCpnt){
			TmpSCpnt = (Scsi_Cmnd *)HPT_SCP(TmpSCpnt)->ptr;
		}
		HPT_SCP(TmpSCpnt)->ptr = HPT_SCP(SCpnt)->ptr;
		if(*SC_Q == SCpnt)
			*SC_Q = TmpSCpnt;
	}
}

void ScsiDone(Scsi_Cmnd *SCpnt)
{
	SCpnt->scsi_done(SCpnt);
	Release_SC(&CURRENT_Q,SCpnt);
    CURRCMDs--;

	if (CURRCMDs == 0)
	{
		if(atomic_read(&DPC_Request_Nums) == 0)
		    Check_Idle_Call();
	}
}

static int scsicmd_buf_get(Scsi_Cmnd *cmd, void **pbuf)
{
	unsigned int buflen;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
	struct scatterlist *sg;
	sg = scsi_sglist(cmd);
	*pbuf = kmap_atomic(HPT_SG_PAGE(sg), OS_KMAP_TYPE) + sg->offset;
	buflen = sg->length;
#else 
	if (cmd->use_sg) {
		struct scatterlist *sg = (struct scatterlist *) cmd->request_buffer;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		*pbuf = kmap_atomic(HPT_SG_PAGE(sg), OS_KMAP_TYPE) + sg->offset;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
		if (sg->page)
			sg->address = kmap_atomic(sg->page, OS_KMAP_TYPE) + sg->offset;
		*pbuf = sg->address;
#else 
		*pbuf = sg->address;
#endif
		buflen = sg->length;
	} else {
		*pbuf = cmd->request_buffer;
		buflen = cmd->request_bufflen;
	}
#endif
	return buflen;
}

static inline void scsicmd_buf_put(struct scsi_cmnd *cmd, void *buf)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
	struct scatterlist *sg;
	sg = scsi_sglist(cmd);
	kunmap_atomic((char *)buf - sg->offset, OS_KMAP_TYPE);
#else 

	if (cmd->use_sg) {
		struct scatterlist *sg = (struct scatterlist *) cmd->request_buffer;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		kunmap_atomic((char *)buf - sg->offset, OS_KMAP_TYPE);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
		if (sg->page)
			kunmap_atomic((char *)buf - sg->offset, OS_KMAP_TYPE);
#endif
	}

#endif
}

static unsigned int fill_msense_rw_recovery(PVDevice pVDev, u8 *p, u32 output_len, u32 bufflen)
{
	const u8  page[] = {0x1, 10, 0xC0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	u32 buf_len = output_len + sizeof(page);

	if (buf_len <= bufflen) {
		memcpy(p, page, sizeof(page));
	}
	return sizeof(page);
}

static unsigned int fill_msense_caching(PVDevice pVDev, u8 *p, u32 output_len, u32 bufflen)
{
	u8 page[20] = {0x8, 18, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	/*page[2] bit meaning:
		0:      write cache disable, write through
		1<<2,   write cache enable
	*/
	u32 buf_len = output_len + sizeof(page);


	if (buf_len <= bufflen) {
		memcpy(p, page, sizeof(page));
	}
	return sizeof(page);
}

static unsigned int fill_msense_ctlmode(u8 *p, u32 output_len, u32 bufflen)
{
	const u8 page[12] = {0xa, 10, 2, 0, 0, 0, 0, 0, 0xff, 0xff, 0, 30};
	u32 buf_len = output_len + sizeof(page);

	if (buf_len <= bufflen) {
		memcpy(p, page, sizeof(page));
	}
	return sizeof(page);
}

#define RW_RECOVERY_MPAGE 0x1
#define CACHE_MPAGE 0x8
#define CONTROL_MPAGE 0xa
#define ALL_MPAGES 0x3f
static void do_mode_sense(PVDevice pVDev, Scsi_Cmnd *SCpnt)
{
	u8 *cdb = SCpnt->cmnd, *p;
	unsigned int page_control, output_len;
	int opcode = cdb[2] & 0x3f;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
	int bufflen = scsi_bufflen(SCpnt);
#else 
	int bufflen = SCpnt->request_bufflen;
#endif

	/* we only support current values */
	page_control = cdb[2] >> 6;
	switch (page_control) {
		case 0: /*current value*/
			break;
		case 3:  /*saved value*/
		case 1:  /*change value*/
		case 2:  /*default value*/
		default:
invalid:
			SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
			return;
	}

	output_len = (cdb[0] == MODE_SENSE)? 4 : 8;

	if (scsicmd_buf_get(SCpnt, (void **)&p)<output_len) {
		scsicmd_buf_put(SCpnt, p);
		goto invalid;
	}

	memset(p, 0, bufflen);

	switch(opcode) {
	case RW_RECOVERY_MPAGE:
		output_len += fill_msense_rw_recovery(pVDev, p+output_len, output_len,  bufflen);
		break;

	case CACHE_MPAGE:
		output_len += fill_msense_caching(pVDev, p+output_len, output_len, bufflen);
		break;

	case CONTROL_MPAGE:
		output_len += fill_msense_ctlmode(p+output_len, output_len, bufflen);
		break;

	case ALL_MPAGES:
		output_len += fill_msense_rw_recovery(pVDev, p+output_len, output_len,  bufflen);
		output_len += fill_msense_caching(pVDev, p+output_len, output_len, bufflen);
		output_len += fill_msense_ctlmode(p+output_len, output_len, bufflen);
		break;

	default:        /* invalid page code */
		scsicmd_buf_put(SCpnt, p);
		goto invalid;
	}

	if (cdb[0] == MODE_SENSE) {
		output_len--;
		p[0] = output_len;
	} else {
		output_len -= 2;
		p[0] = output_len >> 8;
		p[1] = output_len;
	}

	scsicmd_buf_put(SCpnt, p);
	SCpnt->result = (DID_OK<<16);
}


/****************************************************************
 *	Name:	OsSendCommand
 *	Description:	execute a queued command from current_Q.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		viod
 ****************************************************************/
void HPTLIBAPI OsSendCommand (Scsi_Cmnd *SCpnt)
{
	PCommand pCmd;
	PVDevice pVDev = gVBus[sc_channel(SCpnt)].pVDevice[sc_target(SCpnt)];

    Queue_SC(&CURRENT_Q,SCpnt);
	CURRCMDs++;

	switch(SCpnt->cmnd[0]) {
	case START_STOP:
		SCpnt->result = (DID_OK<<16);
        break;

	case INQUIRY:
		{
			PINQUIRYDATA pInquiryData;
			if (scsicmd_buf_get(SCpnt, (void **)&pInquiryData)<36) {
				scsicmd_buf_put(SCpnt, pInquiryData);
				SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
				break;
			}
			ZeroMemory(pInquiryData, 36);
			SetInquiryData(pInquiryData, pVDev);
			scsicmd_buf_put(SCpnt, pInquiryData);
			SCpnt->result = (DID_OK<<16);
			break;
		}
	case READ_CAPACITY:
		{
			UCHAR *rbuf;
			UINT cap;

			if (scsicmd_buf_get(SCpnt, (void **)&rbuf)<8) {
				scsicmd_buf_put(SCpnt, rbuf);
				SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
				break;
			}

			cap = pVDev->VDeviceCapacity - 1;
			rbuf[0] = (u8)(cap>>24);
			rbuf[1] = (u8)(cap>>16);
			rbuf[2] = (u8)(cap>>8);
			rbuf[3] = (u8)cap;
			rbuf[4] = 0;
			rbuf[5] = 0;
			rbuf[6] = 2;
			rbuf[7] = 0;
			scsicmd_buf_put(SCpnt, rbuf);
			SCpnt->result = (DID_OK<<16);
			break;
		}

	case READ_6:
	case WRITE_6:
	case 0x13:      /*SCSIOP_VERIFY6*/
	case READ_10:
	case WRITE_10:
	case VERIFY:
	case 0x88: /* READ_16 */
	case 0x8a: /* WRITE_16 */
	case 0x8f: /* VERIFY_16 */
		{						
			pCmd = AllocateCommand();
			HPT_ASSERT(pCmd);

			switch (SCpnt->cmnd[0]) {
			case READ_6:
			case WRITE_6:
			case 0x13:      /*SCSIOP_VERIFY6*/
				pCmd->uCmd.Ide.Lba = ((ULONG)SCpnt->cmnd[1] << 16) | ((ULONG)SCpnt->cmnd[2] << 8) | (ULONG)SCpnt->cmnd[3];
				pCmd->uCmd.Ide.nSectors = SCpnt->cmnd[4];
				break;
			case 0x88:
			case 0x8a:
			case 0x8f:      /*SCSIOP_VERIFY13*/
				{
					u64 block =
						((u64)SCpnt->cmnd[2]<<56) |
						((u64)SCpnt->cmnd[3]<<48) |
						((u64)SCpnt->cmnd[4]<<40) |
						((u64)SCpnt->cmnd[5]<<32) |
						((u64)SCpnt->cmnd[6]<<24) |
						((u64)SCpnt->cmnd[7]<<16) |
						((u64)SCpnt->cmnd[8]<<8) |
						((u64)SCpnt->cmnd[9]);
					pCmd->uCmd.Ide.Lba = block;
					pCmd->uCmd.Ide.nSectors = (u16)SCpnt->cmnd[13] | ((u16)SCpnt->cmnd[12]<<8);
				}
				break;				

			default:
			    pCmd->uCmd.Ide.Lba = (ULONG)SCpnt->cmnd[5] | ((ULONG)SCpnt->cmnd[4] << 8) |
			         ((ULONG)SCpnt->cmnd[3] << 16) | ((ULONG)SCpnt->cmnd[2] << 24);
			    pCmd->uCmd.Ide.nSectors = (USHORT)SCpnt->cmnd[8] | ((USHORT)SCpnt->cmnd[7]<<8);
			    break;
			}
		    switch (SCpnt->cmnd[0]) {
		    case READ_10:
			case READ_6:
			case 0x88:
		    	pCmd->uCmd.Ide.Command = IDE_COMMAND_READ;
			    pCmd->cf_data_in = 1;
		    	break;
		    case WRITE_10:
			case WRITE_6:
			case 0x8a:
		    	pCmd->uCmd.Ide.Command = IDE_COMMAND_WRITE;
		    	pCmd->cf_data_out = 1;
		    	break;
		    case VERIFY:
		    case 0x13:      /*SCSIOP_VERIFY6*/
			case 0x8f:
		    	pCmd->uCmd.Ide.Command = IDE_COMMAND_VERIFY;
		    	break;
		    }
            KdPrintI(("Os send Command: SCpnt= 0x%x, pVDev=0x%x cmd=%x nSectors=%d lba=%lx\n",SCpnt, pVDev, pCmd->uCmd.Ide.Command, pCmd->uCmd.Ide.nSectors, pCmd->uCmd.Ide.Lba));
            pCmd->pOrgCommand = SCpnt;
			pCmd->pVDevice = pVDev;
			pCmd->pfnBuildSgl = fOsBuildSgl;
			pCmd->pfnCompletion = fOsCommandDone;
			HPT_SCP(SCpnt)->mapped = 0;
		    pVDev->pfnSendCommand(pCmd);
		    goto out;
		}

	case TEST_UNIT_READY:
		SCpnt->result = (DID_OK<<16);
        break;

	case MODE_SENSE:
	case MODE_SENSE_10:
		do_mode_sense(pVDev, SCpnt);
		break;

	default:
		SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
        break;
    }
	ScsiDone(SCpnt);
out:
	CheckPendingCall();
}

/****************************************************************
 *	Name:	hpt3xx_QueueCommand
 *	Description:	Process a queued command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *					done  - Pointer to done function to call.
 *	Returns:		Status code.
 ****************************************************************/
int hpt3xx_QueueCommand (Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
    unsigned long flags;
    PVDevice pVDev;

    KdPrintE(("\nhpt3xx_QueueCommand() %d/%d/%d cdb=(%x-%x-%x)\n",
        sc_channel(SCpnt), sc_target(SCpnt), sc_lun(SCpnt),
	    	*((PULONG)&SCpnt->cmnd), *((PULONG)&SCpnt->cmnd+1), *((PULONG)&SCpnt->cmnd+2)));

	/*
	 * scsi_remove_host issues this cmd, but we have already flushed before and free all mem.
	 */
	if (SCpnt->cmnd[0] == SYNCHRONIZE_CACHE)  {
		SCpnt->result = DID_OK<<16;
		done(SCpnt);
		return 0;
	}

    SCpnt->scsi_done = (void *)done;

    /*
     * start the command. Use spinlock to avoid sending more than one 
     * command at same time
     */
	spin_lock_irqsave(&driver_lock, flags);

	if (sc_channel(SCpnt)>=MAX_VBUS || sc_target(SCpnt)>=MAX_DEVICES || sc_lun(SCpnt))
		pVDev = 0;
	else
		pVDev = gVBus[sc_channel(SCpnt)].pVDevice[sc_target(SCpnt)];
	if (!pVDev) {
		SCpnt->result = (DID_BAD_TARGET<<16);
		SCpnt->scsi_done(SCpnt);
		goto out;
	}
	if ((CURRCMDs == 0) && (atomic_read(&DPC_Request_Nums) == 0))
		Check_Idle_Call();
	if(mWaitingForIdle(gVBus))
	{
		Queue_SC(&PENDING_Q,SCpnt);
		goto out;
	}

    OsSendCommand(SCpnt);
out:
	spin_unlock_irqrestore (&driver_lock, flags);
	KdPrintE(("----- Leaving hpt3xx_queueCommand() -----\n"));
	return 0;
}

/****************************************************************
 *	Name:	internal_done
 *	Description:	Done handler for non-queued commands
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Nothing.
 ****************************************************************/
static void internal_done (Scsi_Cmnd * SCpnt)
{
	HPT_SCP(SCpnt)->status++;
}

/****************************************************************
 *	Name:	hpt3xx_Command
 *	Description:	Process a command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Status code.
 ****************************************************************/
int hpt3xx_Command (Scsi_Cmnd *SCpnt)
{
	unsigned long timeout;
	
	HPT_SCP(SCpnt)->status = 0;
	hpt3xx_QueueCommand (SCpnt, internal_done);

	timeout = jiffies + 60 * HZ; /* should be longer than hard-reset time */
	while (!HPT_SCP(SCpnt)->status && time_before(jiffies, timeout)) schedule();

	if (!HPT_SCP(SCpnt)->status) SCpnt->result = (DID_ERROR<<16);
	return SCpnt->result;
}


/****************************************************************
 *	Name:	hpt3xx_Detect
 *	Description:	Detect and initialize our boards.
 *	Parameters:		tpnt - Pointer to SCSI host template structure.
 *	Returns:		Number of adapters installed.
 ****************************************************************/
int hpt3xx_Detect (Scsi_Host_Template *tpnt)
{
	PChipInstance pChip;
	PChannel  pChan;
	struct pci_dev *pPciDev;
	int i,irqnum;

	printk("%s\nVersion %s, Compiled %s %s\n", 
		DRIVER_NAME, DRIVER_VERSION,  __DATE__, __TIME__);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (!pci_present()) {
		printk(KERN_ERR "PCI BIOS not present\n");
		return 0;
	}
#endif
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	spin_unlock_irq (&io_request_lock);
#endif

	/* setup command blocks */
	pCommandBlocks = VMALLOC(sizeof(struct _Command) * MAX_COMMAND_BLOCKS_FOR_EACH_VBUS);
	if (!pCommandBlocks) {
		hpt_printk(("insufficient memory\n"));
		return 0;
	}
	for (i=0; i<MAX_COMMAND_BLOCKS_FOR_EACH_VBUS; i++)
		FreeCommand(&pCommandBlocks[i]);
	
	if (Initialize()==0) {
		if(gVBus[0].nInstances == 0)
		    return 0;
	}

	/* Allocate a Noncached Extension to use for scatter/gather list */
	for (pChan=pChanStart; pChan<pChanEnd; pChan++) {
		i = pChan-pChanStart;
        SglBuffer[i] = GLOBAL_DATA_ALLOC(MAX_SG_DESCRIPTORS * sizeof(SCAT_GATH) + 16);
		if (!SglBuffer[i]) {
			hpt_printk(("can't alloc SG buffer, exit\n"));
			VFREE(pCommandBlocks);
			return 0;
		}
		pChan->pSgTable = (FPSCAT_GATH)(((ULONG_PTR)SglBuffer[i]+15) & ~(ULONG_PTR)0xF);
		pChan->SgPhysicalAddr = fOsPhysicalAddress(pChan->pSgTable);
	}


	/*
	 * register a virtual host.
	 */
	hpt_vhost = scsi_register(tpnt, sizeof(struct hpt3xx_HOSTDATA));
	hpt_vhost->io_port		= 0;
	hpt_vhost->n_io_port	= 0;
	hpt_vhost->max_channel	= 0;  /* channel <= max_channel */
	hpt_vhost->max_id		= MAX_DEVICES;  /* id < max_id */
	hpt_vhost->max_lun      = 1;  /* lun < max_lun */

	PENDING_Q	            = 0;
	CURRENT_Q	            = 0;
	CURRCMDs		        = 0;
	HOSTDATA(shpnt)->status = FALSE;

	/* 
	 * search for HPT374 chip 
	 */
	pPciDev = NULL;
	i = 0;
	while ((pPciDev=HPT_FIND_PCI_DEVICE(HPT_VENDORID, HPT374_DEVICEID, pPciDev))){
		if (i >= MAX_CHIPS) break;
		for(pChip = gChipInstances; pChip < pChipEnd; pChip++)
			if((pChip->pci_bus == pPciDev->bus->number) && 
				(((pChip->pci_dev<<3) | pChip->pci_func) == pPciDev->devfn))
				break;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
		if (pci_enable_device(pPciDev) < 0) {
				printk("%s: can't enable PCI device, skipping\n",
					DRIVER_NAME);
				continue;
			}
#endif
		HOSTDATA(shpnt)->irq[pChip-gChipInstances] = (UCHAR)pPciDev->irq;
		HOSTDATA(shpnt)->pci_dev = pPciDev;
		pci_set_master(pPciDev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
		if (pci_set_dma_mask(pPciDev, 0xffffffffUL)) {
			hpt_printk(("HPT374: failed to set DMA mask\n"));
			continue;
		}
#endif
		scsi_set_pci_device(hpt_vhost, pPciDev);
		i++;
	}

	if (i!=pChipEnd-gChipInstances) {
		hpt_printk(("PCI enumerating error\n"));
		return 0;
	}

	/*
	 * set up irq handler
	 */	
	memset(irq_map, 0, sizeof(irq_map));
	for(pChip = gChipInstances; pChip < pChipEnd; pChip++) {

		i = pChip-gChipInstances;
		irqnum = HOSTDATA(shpnt)->irq[i];
		while(i) {
			if (irq_map[--i]==irqnum) goto skip_it;
		}
		irq_map[pChip-gChipInstances] = irqnum;
		if (request_irq(irqnum, Irq_Handler, HPT_SA_SHIRQ, PROC_DIR_NAME, gChipInstances) < 0) {
				hpt_printk (("Unable to allocate IRQ\n"));
				return -1;
		}
		skip_it:;
	}

	/* launch worker thread */

#ifndef __BOOT_KERNEL_BOOT
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
    register_reboot_notifier(&hpt_notifier);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	spin_lock_irq (&io_request_lock);
#endif

	return 1;
}

/**************************************************************************
 *    The reset function will reset the SCSI bus and abort any executing
 *    commands. 
 *
 *	Parameters:	   SCpnt - Pointer to SCSI command structure.
 *				   flags - Flags about the reset command
 *  Returns:       DID_RESET in cmd.host_byte of aborted command(s) 
 *  Note:          Resetting the bus always succeeds - is has to, otherwise the
 *                 kernel will panic! Try a surgical technique - sending a BUS
 *                 DEVICE RESET message - on the offending target before pulling
 *                 the SCSI bus reset line.
 **************************************************************************/
int hpt3xx_Reset (Scsi_Cmnd *SCpnt)
{
	/* The Scsi_Cmnd  is passed so that the reset routine can figure
     * out which host adapter should be reset, and also which command
     * within the command block was responsible for the reset in
     * the first place.
	 * Only one virtual host in hpt3xx driver -----hpt_vhost
	 */
	Scsi_Cmnd *SCtmp;
	unsigned long flags;
	
	if (SCpnt==NULL) {
		printk(KERN_WARNING "(scsi?:?:?:?) Reset called with NULL Scsi_Cmnd "
		  "pointer, failing.\n");
		return FAILED;
	}

	hpt_printk(("Reset target: %x lun: %x flags: %x\n",
		sc_target(SCpnt), sc_lun(SCpnt), sc_flags(SCpnt)));
	
	/*
	 * Return immediately if reset is in progress.
	 */
	if (HOSTDATA(shpnt)->status == TRUE) {
		hpt_printk(("reset: exit, already in reset.\n"));
		return FAILED;
	}

	spin_lock_irqsave(&driver_lock,flags);

	HOSTDATA(shpnt)->status = TRUE;
	
	fResetVBus();
	
	HPT_ASSERT(CURRENT_Q==NULL);

   /*
    * Wake-up all awaiting commands with DID_RESET.
    */
	while(PENDING_Q != NULL)
	{
		SCtmp = (Scsi_Cmnd *)HPT_SCP(PENDING_Q)->ptr;
		Release_SC(&PENDING_Q,SCtmp);
		SCtmp->result = (DID_RESET << 16);
		SCtmp->scsi_done(SCtmp);
	}

	/*
	 * Now tell the mid-level code what we did here.  Since
	 * we have restarted all of the outstanding commands,
	 * then report SUCCESS.
	 */
	HOSTDATA(shpnt)->status = FALSE;
	up(&hpt_queue_wt);
	spin_unlock_irqrestore(&driver_lock,flags);
	return SUCCESS;
}

/****************************************************************
 *	Name:	hpt3xx_cleanup
 *	Description:	cleanup routine
 ****************************************************************/
static void hpt3xx_cleanup(void)
{
	PVBus pVBus;
	PVDevice pVDev;
	PChannel pChan;
	int i;
	unsigned long flags;


	spin_lock_irqsave(&driver_lock,flags);

	/* flush all drives */
	FOR_EACH_DEV_ON_ALL_VBUS(pVBus, pVDev, i) {
		fFlushVDev(pVDev);
	}

	VFREE(pCommandBlocks);

	for (pChan=pChanStart; pChan<pChanEnd; pChan++)
		GLOBAL_DATA_FREE(SglBuffer[pChan-pChanStart], MAX_SG_DESCRIPTORS * sizeof(SCAT_GATH)+16);

	spin_unlock_irqrestore(&driver_lock,flags);

	KdPrintI(("\n release resources allocated.\n"));

	/* release IRQ resource */
	for (i=0; i<MAX_CHIPS; i++) {
		if (irq_map[i])
			free_irq (irq_map[i], gChipInstances);
	}
}

/****************************************************************
 *	Name:	hpt3xx_Release
 *	Description:	Release resources allocated for a single each adapter.
 *	Parameters:		pshost - Pointer to SCSI command structure.
 *	Returns:		zero.
 ****************************************************************/
int hpt3xx_Release(struct Scsi_Host *pshost)
{
	hpt3xx_cleanup();
	unregister_reboot_notifier(&hpt_notifier);
    scsi_unregister(pshost);
    return 0;
}

/****************************************************************
 *	Name:	hpt3xx_BiosParam
 *	Description:	Process the biosparam request from the SCSI manager to
 *					return C/H/S data.
 *	Parameters:		disk - Pointer to SCSI disk structure.
 *					dev	 - Major/minor number from kernel.
 *					geom - Pointer to integer array to place geometry data.
 *	Returns:		zero.
 ****************************************************************/
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
int hpt3xx_BiosParam (Scsi_Disk *disk, kdev_t dev, int geom[])
{
	int heads, sectors, cylinders;
	
	/* Default heads (64) & sectors (32) */
	/* Handle extended translation size for logical drives > 1Gb */
	if (disk->capacity >= 0x200000) {
		heads = 255;
		sectors = 63;
	}
	else {
		heads = 64;
		sectors = 32;
	}
	cylinders = disk->capacity / (heads * sectors);
	
	/* return result */
	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;
	
	return 0;
}
#endif

/****************************************************************
 *	Name:	hpt_halt
 *	Description:	shutdown routine
 ****************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
static int hpt_halt(struct notifier_block *nb, ulong event, void *buf)
{
	if (event != SYS_RESTART && event != SYS_HALT && event != SYS_POWER_OFF)
		return NOTIFY_DONE;
		
	hpt3xx_cleanup();

    return NOTIFY_OK;
}
#endif

int HPTLIBAPI fOsBuildSgl(_VBUS_ARG PCommand pCmd, FPSCAT_GATH pSg, int logical)
{
	Scsi_Cmnd *SCpnt = (Scsi_Cmnd *)pCmd->pOrgCommand;
	int       idx;
	int nseg;
	struct scatterlist *sgList;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
	sgList = scsi_sglist(SCpnt);
	nseg = scsi_sg_count(SCpnt);
#else 
	sgList = (struct scatterlist *)SCpnt->request_buffer;
	nseg = SCpnt->use_sg;
#endif

	if (logical) {
		if(nseg){
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)) && defined(CONFIG_HIGHMEM)
			for (idx = 0; idx < nseg; idx++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
				pSg[idx].dSgAddress = (ULONG_PTR)sgList[idx].address;
				if (pSg[idx].dSgAddress) {
					pSg[idx].wSgSize = sgList[idx].length;
				}
				else 
#endif
				{
					/*
					 *  highmem address needs special handling. 
					 *  we store the page in high part of BUS_ADDR and offset in low part
					 *  and map the address to a actual pointer when needed.
					 */
					unsigned int length = sgList[idx].length;
					struct page *page = HPT_SG_PAGE(&sgList[idx]);
					HPT_ASSERT(sizeof(pSg[idx].dSgAddress)>4 && sizeof(void *)==4);
					pSg[idx].dSgAddress = ((u64)(u32)page<<32) | sgList[idx].offset;
					if (sgList[idx].offset + length > PAGE_SIZE) return 0;
					pSg[idx].wSgSize = length;
				}
				pSg[idx].wSgFlag = (idx==nseg-1)? SG_FLAG_EOT : 0;
			}
#else 
			for (idx = 0; idx < nseg; idx++) {
				struct page *page = HPT_SG_PAGE(&sgList[idx]);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
				pSg[idx].dSgAddress = (ULONG_PTR)page_address(page) + sgList[idx].offset;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
				pSg[idx].dSgAddress = (ULONG_PTR)sgList[idx].address;
				if (!pSg[idx].dSgAddress)
					pSg[idx].dSgAddress = (ULONG_PTR)page_address(page) + sgList[idx].offset;
#else 
				pSg[idx].dSgAddress = (ULONG_PTR)sgList[idx].address;
#endif
				pSg[idx].wSgSize = sgList[idx].length;
				pSg[idx].wSgFlag = (idx==nseg-1)? SG_FLAG_EOT : 0;
			}
#endif
		}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)		
		else{
			pSg->dSgAddress = (ULONG_PTR)SCpnt->request_buffer;
			pSg->wSgSize = (USHORT)SCpnt->request_bufflen;
			pSg->wSgFlag = SG_FLAG_EOT;
		}
#endif
	}
	else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
		BUS_ADDR addr, last=0;
		unsigned int length;

		if(!HPT_SCP(SCpnt)->mapped) {
			nseg = scsi_dma_map(SCpnt);
			BUG_ON(nseg < 0);
			if (!nseg)
				return 0;
			HPT_SCP(SCpnt)->mapped = 1;
			HPT_SCP(SCpnt)->sgcnt = nseg;
		}

		pSg--;
		scsi_for_each_sg(SCpnt, sgList, HPT_SCP(SCpnt)->sgcnt, idx) {
			addr = sg_dma_address(sgList);
			length = sg_dma_len(sgList);
			/* merge the sg elements if possible */
			if (idx && last==addr && pSg->wSgSize &&
				pSg->wSgSize+length<=0x10000 && (addr & 0xffffffff)!=0) {
				pSg->wSgSize += length;
				last += length;
			}
			else {
				if (addr & 1) return 0;
				pSg++;
				pSg->dSgAddress = addr;
				pSg->wSgSize = length;
				last = addr + length;
			}
			if (pSg->wSgSize & 1) return 0;
			pSg->wSgFlag = (idx==HPT_SCP(SCpnt)->sgcnt-1)? SG_FLAG_EOT : 0;
		}
#else 
		if (SCpnt->use_sg) {
			BUS_ADDR addr, last=0;
			unsigned int length;
			if (!HPT_SCP(SCpnt)->mapped) {
				HPT_SCP(SCpnt)->sgcnt = pci_map_sg((HOSTDATA(sc_host(SCpnt)))->pci_dev, 
							sgList, SCpnt->use_sg, scsi_to_pci_dma_dir(SCpnt->sc_data_direction));
				HPT_ASSERT(HPT_SCP(SCpnt)->sgcnt <= MAX_SG_DESCRIPTORS);
				HPT_SCP(SCpnt)->mapped = 1;
			}
			pSg--;
			for (idx = 0; idx < HPT_SCP(SCpnt)->sgcnt; idx++) {
				addr = sg_dma_address(&sgList[idx]);
				length = sg_dma_len(&sgList[idx]);
				/* merge the sg elements if possible */
				if (idx && last==addr && pSg->wSgSize &&
					pSg->wSgSize+length<=0x10000 && (addr & 0xffffffff)!=0) {
					pSg->wSgSize += length;
					last += length;
				}
				else {
					if (addr & 1) return FALSE;
					pSg++;
					pSg->dSgAddress = addr;
					pSg->wSgSize = length;
					last = addr + length;
				}
				if (pSg->wSgSize & 1) return FALSE;
				pSg->wSgFlag = (idx==HPT_SCP(SCpnt)->sgcnt-1)? SG_FLAG_EOT : 0;
			}
		}
		else {
			if (!HPT_SCP(SCpnt)->mapped)  {
				HPT_SCP(SCpnt)->dma_handle = pci_map_single(
					(HOSTDATA(sc_host(SCpnt)))->pci_dev, 
					SCpnt->request_buffer,
					SCpnt->request_bufflen,
					scsi_to_pci_dma_dir(SCpnt->sc_data_direction));
				HPT_SCP(SCpnt)->mapped = 1;
			}
			pSg->dSgAddress = HPT_SCP(SCpnt)->dma_handle;
			if ((pSg->dSgAddress & 0x1) ||(!pSg->dSgAddress))return FALSE;
			pSg->wSgSize = (USHORT)SCpnt->request_bufflen;
			if (pSg->wSgSize & 0x1) return FALSE;
			pSg->wSgFlag = SG_FLAG_EOT;
		}
#endif
	}
	return TRUE;
}

void HPTLIBAPI fOsCommandDone(PCommand pCmd)
{
	Scsi_Cmnd *SCpnt = (Scsi_Cmnd *)pCmd->pOrgCommand;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
	if (HPT_SCP(SCpnt)->mapped) { 
		scsi_dma_unmap(SCpnt);
	}
#else 
	if (HPT_SCP(SCpnt)->mapped) {
		if (SCpnt->use_sg)
			pci_unmap_sg(
				HOSTDATA(sc_host(SCpnt))->pci_dev,
				(struct scatterlist *)SCpnt->request_buffer,
				SCpnt->use_sg,
				scsi_to_pci_dma_dir(SCpnt->sc_data_direction)
			);
		else
				pci_unmap_single(
				HOSTDATA(sc_host(SCpnt))->pci_dev,
				HPT_SCP(SCpnt)->dma_handle,
				SCpnt->request_bufflen,
				scsi_to_pci_dma_dir(SCpnt->sc_data_direction)
			);
	}
#endif

	switch(pCmd->Result) {
	case RETURN_SUCCESS:
		KdPrint(("RETURN_SUCCESS\n"));
		SCpnt->result = (DID_OK<<16);
		break;
	case RETURN_BAD_DEVICE:
		KdPrint(("RETURN_BAD_DEVICE"));
		SCpnt->result = (DID_BAD_TARGET<<16);
		break;
	case RETURN_DEVICE_BUSY:
		KdPrint(("RETURN_DEVICE_BUSY"));
		SCpnt->result = (DID_BUS_BUSY<<16);
		break;
	case RETURN_RETRY:
		KdPrint(("RETURN_RETRY"));
		SCpnt->result = (DID_BUS_BUSY<<16);
		break;
	case RETURN_SELECTION_TIMEOUT:
		KdPrint(("RETURN_SELECTION_TIMEOUT"));
		SCpnt->result = (DID_NO_CONNECT<<16);
		break;
	case RETURN_BUS_RESET:
		KdPrint(("RETURN_BUS_RESET"));
		SCpnt->result = (DID_RESET<<16);
		break;
	default:
		KdPrint(("RETURN_ERROR"));
		SCpnt->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
		break;
	}
	FreeCommand(pCmd);
	ScsiDone(SCpnt);
}

ULONG HPTLIBAPI fAsMapLbaToCHS(ULONG Lba, WORD sectorXhead, BYTE sector_per_track)
{
	/* LBA = ((C * heads + H) * sectors_per_track ) + S - 1 */
	ULONG C, H, S;
	USHORT heads = sectorXhead / sector_per_track;
	S = Lba % sector_per_track + 1;
	H = Lba / sector_per_track % heads;
	C = Lba / sector_per_track / heads;
	/*
	 * return 28bit address.
	 */
	return S | C<<8 | H<<24;
}

void HPTLIBAPI OsSetDeviceTable(PDevice pDevice, PIDENTIFY_DATA pIdentify)
{
    MemoryCopy(&pDevice->IdentifyData, pIdentify, sizeof(IDENTIFY_DATA2));
}

static void SetInquiryData(PINQUIRYDATA inquiryData, PVDevice pVDev)
{
	int i;		
	inquiryData->DeviceType = TYPE_DISK; /*DIRECT_ACCESS_DEVICE*/
	inquiryData->AdditionalLength = (UCHAR)(sizeof(INQUIRYDATA) - 5);
#ifndef SERIAL_CMDS
	inquiryData->CommandQueue = 1;
#endif

	switch(pVDev->VDeviceType) {
	case VD_SINGLE_DISK:
	case VD_ATAPI:
	case VD_REMOVABLE:

		if ((pVDev->u.disk.df_removable_drive) || 
			(pVDev->u.disk.IdentifyData.GeneralConfiguration & cpu_to_le16(0x80)))
			inquiryData->RemovableMedia = 1;

		for (i = 0; i < 20; i += 2)
		{				
			inquiryData->VendorId[i] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.ModelNumber)[i + 1];
			inquiryData->VendorId[i+1] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.ModelNumber)[i];

		}

		for (i = 0; i < 4; i++) inquiryData->ProductId[12+i] = ' ';

		for (i = 0; i < 4; i += 2)
		{				
			inquiryData->ProductRevisionLevel[i] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.FirmwareRevision)[i+1];
			inquiryData->ProductRevisionLevel[i+1] 	= ((PUCHAR)pVDev->u.disk.IdentifyData.FirmwareRevision)[i];
		}
		break;
	default:
		memcpy(&inquiryData->VendorId, "HPT3xx  ", 8);
		memcpy(&inquiryData->ProductRevisionLevel, "3.00", 4);
	    break;
	}
}

/***************************************************************************
 * Function:     hpt_worker_thread
 * Description:  Do background rebuilding. Execute in kernel thread context.
 * Returns:      None
 ***************************************************************************/

/***************************************************************************
 * Function:     hpt_queue_dpc
 * Description:  Queue a DPC called by worker thread. It may be called in an ISR.
 * Returns:      None
 ***************************************************************************/
int hpt_queue_dpc(HPT_DPC dpc, void *arg, UCHAR flags)
{
	int p;
	unsigned long Lock_Flag;
	/*
	 * driver_lock shall be locked before call this function
	 */
	spin_lock_irqsave(&DpcQueueLock, Lock_Flag);	 
	p = (DpcQueue_Last + 1) % MAX_DPC;
	if (p==DpcQueue_First) {
		KdPrintW(("DPC Queue full!\n"));
		return -1;
	}
	DpcQueue[DpcQueue_Last].dpc = dpc;
	DpcQueue[DpcQueue_Last].arg = arg;
	DpcQueue[DpcQueue_Last].flags = flags;
	DpcQueue_Last = p;
	up(hpt_queue_ready);
	spin_unlock_irqrestore(&DpcQueueLock, Lock_Flag);	
	return 0;
}

void hpt_refresh_logical_devices(void) {}

/*
 * Host template
 */
static Scsi_Host_Template driver_template = {
	name:				     PROC_DIR_NAME,
	detect:				     hpt3xx_Detect,
	release:			     hpt3xx_Release,
	queuecommand:		     hpt3xx_QueueCommand,
	eh_device_reset_handler: hpt3xx_Reset,
	eh_bus_reset_handler:    hpt3xx_Reset,
	can_queue:			     MAX_QUEUE_COMM,
	sg_tablesize:		     16,
	/* larger cmd_per_lun will cause Edma queue full, and hang up RAID5 case. */
	cmd_per_lun:		     8,
	unchecked_isa_dma:	     0,
	emulated:       	     0,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)   /* 2.2.x */
	use_new_eh_code:	     1,
	proc_dir:			     &proc_scsi_hpt3xx,
	command:			     hpt3xx_Command,		
	proc_info:			     hpt3xx_ProcInfo,
	bios_param:			     hpt3xx_BiosParam,
	use_clustering: 	     ENABLE_CLUSTERING, 
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0) /* 2.4.x */
	use_new_eh_code:	     1,
	proc_name:			     PROC_DIR_NAME,
	proc_info:			     hpt3xx_ProcInfo,
	command:			     hpt3xx_Command,		
	bios_param:			     hpt3xx_BiosParam,
	max_sectors:		     128, 
	use_clustering: 	     ENABLE_CLUSTERING, 
#else /* 2.6.x */
	proc_name:			     PROC_DIR_NAME,
	proc_info:			     hpt3xx_ProcInfo26,
	max_sectors:		     128, 
	use_clustering: 	     DISABLE_CLUSTERING, 
#endif
	this_id:			     -1
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)

#include "scsi_module.c"

EXPORT_NO_SYMBOLS;
/* PCI space access */
UCHAR HPTLIBAPI pcicfg_read_byte (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	UCHAR v;
	if (pcibios_read_config_byte(bus, (dev<<3)|func, reg, &v)) return 0xff;
	return v;
}
WORD HPTLIBAPI pcicfg_read_word(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	WORD v;
	if (pcibios_read_config_word(bus, (dev<<3)|func, reg, &v)) return 0xffff;
	return v;
}
DWORD HPTLIBAPI pcicfg_read_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	DWORD v;
	if (pcibios_read_config_dword(bus, (dev<<3)|func, reg, (unsigned int *)&v)) return 0xffffffff;
	return v;
}
void HPTLIBAPI pcicfg_write_byte (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, UCHAR v)
{
	pcibios_write_config_byte(bus, (dev<<3)|func, reg, v);
}
void HPTLIBAPI pcicfg_write_word(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, WORD v)
{
	pcibios_write_config_word(bus, (dev<<3)|func, reg, v);
}
void HPTLIBAPI pcicfg_write_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, DWORD v)
{
	pcibios_write_config_dword(bus, (dev<<3)|func, reg, v);
}

#else /* 2.6.x */

/*scsi_module.c will be deprecated since 2.6.x*/
static int __init init_this_scsi_driver(void)
{
	struct scsi_host_template *sht = &driver_template;
	struct Scsi_Host *shost;
	struct list_head *l;
	int error;

	if (!sht->release) {
		printk(KERN_ERR
			"scsi HBA driver %s didn't set a release method.\n",
			sht->name);
		return -EINVAL;
	}

	sht->module = THIS_MODULE;
	INIT_LIST_HEAD(&sht->legacy_hosts);

	sht->detect(sht);
	if (list_empty(&sht->legacy_hosts))
		return -ENODEV;

	list_for_each_entry(shost, &sht->legacy_hosts, sht_legacy_list) {
		error = scsi_add_host(shost, &((HOSTDATA(shost))->pci_dev->dev));
		if (error)
			goto fail;
		scsi_scan_host(shost);
	}
	return 0;
 fail:
	l = &shost->sht_legacy_list;
	while ((l = l->prev) != &sht->legacy_hosts)
		scsi_remove_host(list_entry(l, struct Scsi_Host, sht_legacy_list));
	return error;
}

static void __exit exit_this_scsi_driver(void)
{
	struct scsi_host_template *sht = &driver_template;
	struct Scsi_Host *shost, *s;

	list_for_each_entry(shost, &sht->legacy_hosts, sht_legacy_list)
		scsi_remove_host(shost);
	list_for_each_entry_safe(shost, s, &sht->legacy_hosts, sht_legacy_list)
		sht->release(shost);

	if (list_empty(&sht->legacy_hosts))
		return;

	printk(KERN_WARNING "%s did not call scsi_unregister\n", sht->name);
	dump_stack();

	list_for_each_entry_safe(shost, s, &sht->legacy_hosts, sht_legacy_list)
		scsi_unregister(shost);
}

module_init(init_this_scsi_driver);
module_exit(exit_this_scsi_driver);

UCHAR HPTLIBAPI pcicfg_read_byte (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	UCHAR v;
	struct pci_dev *pPciDev = HPT_FIND_SLOT_DEVICE(bus, PCI_DEVFN(dev,func));
	if (pPciDev == NULL)
		return 0xff;
	if (pci_read_config_byte(pPciDev, reg, &v)) return 0xff;
	return v;
}
WORD HPTLIBAPI pcicfg_read_word(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	WORD v;
	struct pci_dev *pPciDev = HPT_FIND_SLOT_DEVICE(bus, PCI_DEVFN(dev,func));
	if (pPciDev == NULL)
		return 0xffff;
	if (pci_read_config_word(pPciDev, reg, &v)) return 0xffff;
	return v;
}
DWORD HPTLIBAPI pcicfg_read_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg)
{
	DWORD v;
	struct pci_dev *pPciDev = HPT_FIND_SLOT_DEVICE(bus, PCI_DEVFN(dev,func));
	if (pPciDev == NULL)
		return 0xffffffff;
	if (pci_read_config_dword(pPciDev, reg, (unsigned int *)&v)) return 0xffffffff;
	return v;
}

void HPTLIBAPI pcicfg_write_byte(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, UCHAR v)
{
	struct pci_dev *pPciDev = HPT_FIND_SLOT_DEVICE(bus, PCI_DEVFN(dev,func));
	if (pPciDev == NULL)
		return ;
	pci_write_config_byte(pPciDev,reg, v);
}
void HPTLIBAPI pcicfg_write_word(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, WORD v)
{
	struct pci_dev *pPciDev = HPT_FIND_SLOT_DEVICE(bus, PCI_DEVFN(dev,func));
	if (pPciDev == NULL)
		return ;
	pci_write_config_word(pPciDev,reg, v);
}
void HPTLIBAPI pcicfg_write_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, DWORD v)
{
	struct pci_dev *pPciDev = HPT_FIND_SLOT_DEVICE(bus, PCI_DEVFN(dev,func));
	if (pPciDev == NULL)
		return ;
	pci_write_config_dword(pPciDev,reg, v);
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
MODULE_LICENSE("Proprietary");
#endif


BYTE hpt_inp(unsigned port) { return inb(port); }
WORD hpt_inpw(unsigned port) { return inw(port); }
DWORD hpt_inpd(unsigned port) { return inl(port); }
void hpt_outp(unsigned port, BYTE value) { outb(value, port); }
void hpt_outpw(unsigned port, WORD value) { outw(value, port); }
void hpt_outpd(unsigned port, DWORD value) { outl(value, port); }
void hpt_repinsw(unsigned ioport, ADDRESS buffer, WORD count) { insw(ioport, buffer, count); }
void hpt_repoutsw(unsigned ioport, ADDRESS buffer, WORD count) { outsw(ioport, buffer, count); }


void HPTLIBAPI StallExec(DWORD x)
{
	(x < MAX_UDELAY_MS*1000)? udelay(x) : mdelay(x/1000);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0) && defined (__x86_64__)
	touch_nmi_watchdog();
#endif
}
ULONG_PTR HPTLIBAPI fOsPhysicalAddress(void *addr) { return virt_to_bus(addr); }
void * HPTLIBAPI os_alloc_page(_VBUS_ARG0) { return VMALLOC(0x1000); }
void * HPTLIBAPI os_alloc_dma_page(_VBUS_ARG0) { return kmalloc(0x1000, GFP_ATOMIC); }
void HPTLIBAPI os_free_page(_VBUS_ARG void *p) { VFREE(p); }
void HPTLIBAPI os_free_dma_page(_VBUS_ARG void *p) { kfree(p); }

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)) && defined(CONFIG_HIGHMEM)
void * HPTLIBAPI os_kmap_ptr(BUS_ADDR addr)
{

	struct page * page = (struct page *)(ULONG_PTR)(addr>>32);
	if (page) 
	
		return (PageHighMem(page)? (char *)kmap_atomic(page, OS_KMAP_TYPE) : (char *)page_address(page))
			+ (addr & 0xffffffff);
	
	else

		return (void *)(ULONG_PTR)addr;

}

void HPTLIBAPI os_kunmap_ptr(void *ptr)
{
	if ((ULONG_PTR)ptr >= (ULONG_PTR)high_memory)
		kunmap_atomic(ptr, OS_KMAP_TYPE);
}
#else 
void * HPTLIBAPI os_kmap_ptr(BUS_ADDR addr) { return (void *)(ULONG_PTR)addr; }
void HPTLIBAPI os_kunmap_ptr(void *ptr) {}
#endif

void HPTLIBAPI DoXor1(ULONG *p0, ULONG *p1, ULONG *p2, UINT nBytes)
{
    UINT i;
	for (i = 0; i < nBytes / 4; i++) *p0++ = *p1++ ^ *p2++;
}

void HPTLIBAPI DoXor2(ULONG *p0, ULONG *p2, UINT nBytes)
{
    UINT i;
	for (i = 0; i < nBytes / 4; i++) *p0++ ^= *p2++;
}

int HPTLIBAPI os_printk(char *fmt, ...)
{
	va_list ap;
	char buf[256];
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	return printk("%s\n", buf);
}

int HPTLIBAPI os_memcmp(const void *cs, const void *ct, unsigned len) { return memcmp(cs, ct, len); }
void HPTLIBAPI os_memcpy(void *to, const void *from, unsigned len) { memcpy(to, from, len); }
void HPTLIBAPI os_memset(void *s, char c, unsigned len) { memset(s, c, len); }
unsigned HPTLIBAPI os_strlen(const char *s) { return strlen(s); }

WORD  HPTLIBAPI CPU_TO_LE16(WORD x)  { return cpu_to_le16(x); }
DWORD HPTLIBAPI CPU_TO_LE32(DWORD x) { return cpu_to_le32(x); }
WORD  HPTLIBAPI LE16_TO_CPU(WORD x)  { return le16_to_cpu(x); }
DWORD HPTLIBAPI LE32_TO_CPU(DWORD x) { return le32_to_cpu(x); }
WORD  HPTLIBAPI CPU_TO_BE16(WORD x)  { return cpu_to_be16(x); }
DWORD HPTLIBAPI CPU_TO_BE32(DWORD x) { return cpu_to_be32(x); }
WORD  HPTLIBAPI BE16_TO_CPU(WORD x)  { return be16_to_cpu(x); }
DWORD HPTLIBAPI BE32_TO_CPU(DWORD x) { return be32_to_cpu(x); }
