/*
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * Copyright (C) 2000-2003 HighPoint Technologies, Inc.
 * oslinux.h  Linux headers for platform dependent part.
 */
#ifndef _OSLINUX_H_
#define _OSLINUX_H_

#define hpt_printk(x) do { printk(PROC_DIR_NAME": "); printk x; } while (0)

extern struct Scsi_Host *hpt_vhost;

typedef struct 
{
	UCHAR		status;     		/* 0 nonbootable; 80h bootable */
	UCHAR      	start_head;
	USHORT     	start_sector;
	UCHAR      	type;
	UCHAR      	end_head;
	USHORT     	end_sector;
	ULONG      	start_abs_sector;
	ULONG      	num_of_sector;
} partition;

typedef struct _INQUIRYDATA {
#ifdef __BIG_ENDIAN_BITFIELD
	UCHAR DeviceTypeQualifier : 3;
	UCHAR DeviceType : 5;
	UCHAR RemovableMedia : 1;
	UCHAR DeviceTypeModifier : 7;
#else 
	UCHAR DeviceType : 5;
	UCHAR DeviceTypeQualifier : 3;
	UCHAR DeviceTypeModifier : 7;
	UCHAR RemovableMedia : 1;
#endif
	UCHAR Versions;
	UCHAR ResponseDataFormat;
	UCHAR AdditionalLength;
	UCHAR Reserved[2];
#ifdef __BIG_ENDIAN_BITFIELD
	UCHAR RelativeAddressing : 1;
	UCHAR Wide32Bit : 1;
	UCHAR Wide16Bit : 1;
	UCHAR Synchronous : 1;
	UCHAR LinkedCommands : 1;
	UCHAR Reserved2 : 1;
	UCHAR CommandQueue : 1;
	UCHAR SoftReset : 1;
#else 
	UCHAR SoftReset : 1;
	UCHAR CommandQueue : 1;
	UCHAR Reserved2 : 1;
	UCHAR LinkedCommands : 1;
	UCHAR Synchronous : 1;
	UCHAR Wide16Bit : 1;
	UCHAR Wide32Bit : 1;
	UCHAR RelativeAddressing : 1;
#endif
	UCHAR VendorId[8];
	UCHAR ProductId[16];
	UCHAR ProductRevisionLevel[4];
	UCHAR VendorSpecific[20];
	UCHAR Reserved3[40];
} INQUIRYDATA, *PINQUIRYDATA;

struct hpt3xx_HOSTDATA {
	Scsi_Cmnd *pending_Q;
		/* pending commands to issue */
	Scsi_Cmnd *current_Q;
		/* current commands on the bus */
	UCHAR commands;		
	    /* number of current commands,this way only allow commands==32 */
	UCHAR status;
		/* host lock */
	UCHAR irq[MAX_CHIPS];
	struct pci_dev *pci_dev;
};
/* access macros for hostdata */
#define HOSTDATA(shpnt)		((struct hpt3xx_HOSTDATA *) &hpt_vhost->hostdata)
#define PENDING_Q           (HOSTDATA(shpnt)->pending_Q)
#define CURRENT_Q		    (HOSTDATA(shpnt)->current_Q)
#define CURRCMDs            (HOSTDATA(shpnt)->commands)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define pci_map_sg(pdev, sg_list, nseg, direction) (nseg)
#define pci_unmap_sg(pdev, sg_list, nseg, direction)
#define pci_map_single(pdev,buffer,size,direction) virt_to_bus(buffer)
#define pci_unmap_single(pdev,bus_addr,size,direction)
#define sg_dma_address(sg) (virt_to_bus((sg)->address))
#define sg_dma_len(sg) ((sg)->length)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define sc_host(sc)    sc->device->host
#define sc_channel(sc) sc->device->channel
#define sc_target(sc)  sc->device->id
#define sc_lun(sc)     sc->device->lun
#define sc_flags(sc)     sc->eh_eflags
#else 
#define sc_host(sc)    sc->host
#define sc_channel(sc) sc->channel
#define sc_target(sc)  sc->target
#define sc_lun(sc)     sc->lun
#define sc_flags(sc)   sc->flags
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define scsi_to_pci_dma_dir(scsi_dir) ((int)(scsi_dir))
#define OS_KMAP_TYPE KM_BIO_SRC_IRQ
#define HPT_FIND_PCI_DEVICE pci_get_device
#else 
#define OS_KMAP_TYPE KM_BH_IRQ
#define HPT_FIND_PCI_DEVICE pci_find_device
#endif

/*verify_area will be deprecated in the following 2.6.11 kernel*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
#define hpt_verify_area(type, addr, size) (!access_ok((type), (addr), (size)))
#else 
#define hpt_verify_area(type, addr, size) (verify_area((type), (addr), (size)))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
#define HPT_SA_SHIRQ		IRQF_SHARED
#else 
#define HPT_SA_SHIRQ		SA_SHIRQ
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
#define HPT_FIND_SLOT_DEVICE pci_get_bus_and_slot
#else 
#define HPT_FIND_SLOT_DEVICE pci_find_slot
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
#define HPT_SG_PAGE(sg)		sg_page(sg)
#else 
#define HPT_SG_PAGE(sg)		(sg)->page
#endif

struct hpt_scsi_pointer {
	char * ptr;
	int status;
	int mapped;
	int sgcnt;
	dma_addr_t dma_handle;
};
#define HPT_SCP(SCpnt) ((struct hpt_scsi_pointer *)&(SCpnt)->SCp)

extern spinlock_t driver_lock;

/*entry.c*/
int HPTLIBAPI fOsBuildSgl(PCommand pCmd, FPSCAT_GATH pSg, int logical);
typedef void (*HPT_DPC)(void*,UCHAR);
int hpt_queue_dpc(HPT_DPC dpc, void *arg, UCHAR flags);
void hpt_rebuild_data_block(PVDevice pArray, UCHAR flags);

#define DUPLICATE      0
#define INITIALIZE     1
#define REBUILD_PARITY 2
#define VERIFY_R         3

#endif
